/****************************************************************************
 *
 * Stack2: Template-Klasse Stack mit statischer Groesse
 *
 *  - Beispielprogramm zum Buch "Objektorientiertes Programmieren in C++",
 *    von Nicolai Josuttis, erschienen bei Addison-Wesley, ISBN 3-89319-637-4
 *
 *  - Abschnitt 3.14.3: Andere Template-Argumente
 *  
 ****************************************************************************/

/****************************************************************************
 *
 * Copyright 1994 by Nicolai Josuttis and Addison-Wesley.
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software for personal
 * and educational use is hereby granted without fee, provided that the above
 * copyright notice appear in all copies and that both that copyright notice
 * and this permission notice appear in supporting documentation, and that
 * the names of Addison-Wesley or the authors are not be used in advertising
 * or publicity pertaining to distribution of the software without specific,
 * written prior permission.  Addison-Wesley and the authors make no
 * representations about the suitability of this software for any purpose.
 * It is provided "as is" without express or implied warranty.
 *
 * ADDISON-WESLEY AND THE AUTHORS DISCLAIM ALL WARRANTIES WITH REGARD TO THIS
 * SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
 * IN NO EVENT SHALL ADDISON-WESLEY OR THE AUTHORS BE LIABLE FOR ANY SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 *
 ****************************************************************************/

// da die meisten Compiler "bool" (noch) nicht koennen (siehe Seite 43):
typedef int bool;
#define true 1
#define false 0



/****************************************************************************
 *
 *	String.h
 *
 ****************************************************************************/

#ifndef _String_h
#define _String_h

// Header-Datei fuer I/O
#include <iostream.h>

class String {

  private:
    char*    cstring;    // Zeichenfolge als dynamischer C-String
    unsigned len;        // Anzahl Zeichen (ohne '\0')
    unsigned size;       // Speicherplatzgroesse von "cstring"

  public:
    // Default- und char*-Konstruktor
    String (const char* = "");

    // Vergleichen von Strings
    friend bool operator== (const String&, const String&);
    friend bool operator!= (const String&, const String&);

    // Hintereinanderhaengen von Strings
    friend String operator+ (const String&, const String&);

    // Ausgabe mit Streams
    void printOn (ostream& = cout) const;

    // Eingabe mit Streams
    void scanFrom (istream& = cin);

    // Aufgrund dynamischer Komponenten
    String (const String&);                   // Copy-Konstruktor
    const String& operator= (const String&);  // Zuweisung
    ~String ();                               // Destruktor
  private:
    /* Konstruktor fuer uninitialisierten String bestimmter Laenge
     * (intern fuer operator+())
     */
    String (unsigned);
};

// Standard-Ausgabeoperator
inline ostream& operator<< (ostream& strm, const String& s)
{
    s.printOn (strm);    // String auf Stream ausgeben
    return strm;         // Stream zurueckliefern
}

// Standard-Eingabeoperator
inline istream& operator>> (istream& strm, String& s)
{
    s.scanFrom (strm);   // String von Stream einlesen
    return strm;         // Stream zurueckliefern
}

/* Operator !=
 *  - als Umsetzung auf Operator == inline implementiert
 */
inline bool operator!= (const String& s1, const String& s2) {
    return !(s1==s2);
}

#endif  /* _String_h */


/****************************************************************************
 *
 *	String.cc / String.cpp
 *
 ****************************************************************************/

// Header-Datei der eigenen Klasse
//#include "String.h"

// C-Header-Dateien fuer String-Funktionen
#include <string.h>
#include <ctype.h>

/* Konstruktor aus C-String (char*)
 *  - Default fuer s: ""
 */

String::String (const char* s)
{
    len = strlen (s);             // Anzahl Zeichen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
    strcpy (cstring, s);          // Zeichen in Speicherplatz kopieren
}

/* Copy-Konstruktor
 */
String::String (const String& s)
{
    len = s.len;                    // Anzahl Zeichen uebernehmen
    size = len + 1;                 // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];      // Speicherplatz anlegen
    strcpy (cstring, s.cstring);    // Zeichen kopieren
}

/* Destruktor
 */
String::~String ()
{
    // Mit new angelegten Speicherplatz wieder freigeben
    delete [] cstring;
}

/* operator =
 *  - Zuweisung
 */
const String& String::operator= (const String& s)
{
    // Zuweisung eines Strings an sich selbst macht nichts
    if (this == &s) {
        return *this;                 // String zurueckliefern    
    }

    len = s.len;             // Anzahl Zeichen uebernehmen

    // falls Platz nicht reicht, vergroessern
    if (size < len + 1) {
        delete [] cstring;            // alten Speicherplatz freigeben
        size = len + 1;               // Anzahl Zeichen plus '\0'
        cstring = new char [size];    // Speicherplatz anlegen
    }

    strcpy (cstring, s.cstring);      // Zeichen kopieren

    return *this;            // geaenderten String zurueckliefern    
}

/* operator ==
 *  - vergleicht zwei Strings
 *  - globale Friend-Funktion, damit automatische
 *    Typumwandlung des ersten Operanden moeglich ist
 */
bool operator== (const String& s1, const String& s2)
{
    return strcmp(s1.cstring,s2.cstring) == 0;
}

/* operator +
 *  - haengt zwei Strings hintereinander
 *  - globale Friend-Funktion, damit automatische
 *    Typumwandlung des ersten Operanden moeglich ist
 */
String operator+ (const String& s1, const String& s2)
{
    // uninitialisierten Summen-String erzeugen
    String sum (s1.len + s2.len + 1);

    // Summen-Zeichenfolge darin initialisieren
    strcpy (sum.cstring, s1.cstring);
    strcat (sum.cstring, s2.cstring);

    // Kopie von Summen-String zurueck
    return sum;
}

/* Konstruktor fuer uninitialisierten String bestimmter Laenge
 *  - intern fuer operator+
 */
String::String (unsigned l)
{
    len = l;                      // Anzahl Zeichen uebernehmen
    size = len + 1;               // Size ist Anzahl Zeichen plus '\0'
    cstring = new char [size];    // Speicherplatz anlegen
}

/* Ausgabe auf Stream
 */
void String::printOn (ostream& strm) const
{
    // Zeichenfolge einfach ausgeben
    strm << cstring;
}

void String::scanFrom (istream& strm)
{
    char c;

    len = 0;         // zunaechst ist der gelesene String leer

    strm >> ws;      // fuehrende Trennzeichen ueberlesen

    /* solange Input-Stream "strm" nach Einlesen
     * eines Zeichens in "c" in Ordnung ist
     */
    while (strm.get(c)) {       // >> wuerde Trennzeichen ueberlesen

        /* Newline schliesst Stringeingabe ab
         *  - '\n' => RETURN
         */
        if (c == '\n') {
            cstring [len] = '\0';
            return;
        }

        /* falls der Platz nicht mehr ausreicht,
         * muss mehr geschaffen werden
         */
        if (len+1 >= size) {
            char* tmp = cstring;       // Zeiger auf alten Speicherplatz
            size += 80;                // 80 Zeichen mehr Platz
            cstring = new char[size];  // neuen Speicherplatz anlegen
            strcpy (cstring, tmp);     // Zeichen kopieren
            delete [] tmp;             // alten Speicherplatz freigeben
        }

        // neues Zeichen eintragen
        cstring [len] = c;
        len ++;
    }

    /* Einlese-Ende durch Fehler oder EOF
     *  - Stringende-Kennzeichen eintragen
     */
    cstring [len] = '\0';
}



/****************************************************************************
 *
 *	Stack.h
 *
 ****************************************************************************/

template <class T, int size>
class Stack {
  private:
    T elems[size];       // Array von size Elementen vom Typ T
    int numElems;        // Anzahl eingetragener Elemente

  public:
    Stack ();            // Konstruktor
    void push (T&);      // Element einkellern
    T pop ();            // Element auskellern
};


/****************************************************************************
 *
 *	Stack.cc / Stack.cpp
 *
 ****************************************************************************/

// Header-Dateien
#include <stdlib.h>
#include <iostream.h>
//#include "Stack.h"

/* Konstruktor
 */
template <class T, int size>
Stack<T,size>::Stack ()
{
    numElems = 0;               // zunaechst keine Elemente
}

template <class T, int size>
void Stack<T,size>::push (T& x)
{
    // Groesse ueberschritten ?
    if (numElems == size) {
        cerr << "Stack ist voll" << endl;
        exit (1);
    }

    elems[numElems++] = x;      // neues Element einkellern
}

template<class T, int size>
T Stack<T,size>::pop ()
{
    // kein Element mehr vorhanden ?
    if (numElems == 0) {
        cerr << "Stack ist leer" << endl;
        exit (1);
    }

    return elems[--numElems];   // Element auskellern
}



/****************************************************************************
 *
 *	stest2.cc / stest2.cpp
 *
 ****************************************************************************/

#include <iostream.h>
//#include "String.h"
//#include "Stack.h"

typedef Stack<String,10> Str10Stack;

main()
{
    Str10Stack strStack;     // Stack fuer Strings (Startgroesse: 10)

    static String worte [5] = {"gehen","zu","scheint","Stack","der"};
    int i;

    cout << "push: " << endl;
    for (i=0; i<5; i++) {
        cout << " " << worte[i] << endl;
	strStack.push(worte[i]);
    }
    
    cout << "\npop: " << endl;
    for (i=0; i<5; i++) {
        cout << " " << strStack.pop();
    }
    cout << endl;
}

